import math
from abc import abstractmethod

import osmnx as ox
import pandas as pd
import quads

from urban_transit_network_analysis.context import GraphAnalysContext
from urban_transit_network_analysis.database.Neo4jConnection import Neo4jConnection
from urban_transit_network_analysis.parser.Parser import BusGraphParser, TrolleyGraphParser,\
    TramGraphParser, MiniBusGraphParser
"""
    Классы с наследственной структурой которые занимаются работой с бд для каждого типа сети 
"""


class GraphDBManager:
    def __init__(self, graph_analis_context: GraphAnalysContext):
        self.connection = Neo4jConnection()
        self.city_name = graph_analis_context.city_name
        self.enrichDBParameters(graph_analis_context)
        self.db_graph_parameters = graph_analis_context.neo4j_DB_graph_parameters

    def enrichDBParameters(self, graph_analis_context: GraphAnalysContext):
        graph_analis_context.neo4j_DB_graph_parameters.city_name = self.city_name
        graph_analis_context.neo4j_DB_graph_parameters.node_geometry_identity = self.node_geometry_identity()
        graph_analis_context.neo4j_DB_graph_parameters.main_rels_name = self.get_main_rels_name()
        graph_analis_context.neo4j_DB_graph_parameters.main_node_name = self.get_main_node_name()
        graph_analis_context.neo4j_DB_graph_parameters.weight = self.get_weight()

    @abstractmethod
    def get_graph(self):
        pass

    @abstractmethod
    def get_weight(self):
        pass

    @abstractmethod
    def get_main_node_name(self):
        pass

    @abstractmethod
    def get_main_rels_name(self):
        pass

    @abstractmethod
    def node_geometry_identity(self):
        pass


class TwoTypeNodeDBManager(GraphDBManager):
    def __init__(self, graph_analis_context: GraphAnalysContext):
        super().__init__(graph_analis_context)
        graph_analis_context.neo4j_DB_graph_parameters.secondary_node_name = self.get_second_node_name()
        graph_analis_context.neo4j_DB_graph_parameters.secondary_rels_name = self.get_second_rels_name()
        self.db_graph_parameters.secondary_node_name = self.get_second_node_name()
        self.db_graph_parameters.secondary_rels_name = self.get_second_rels_name()

    def update_db(self):
        (first_nodes, first_relationships, second_nodes, second_relationships) = self.get_graph()
        if first_nodes is None and first_relationships is None and second_nodes is None and second_relationships is None:
            print("Graph for", self.city_name, "is empty!")
            return
        self.connection.execute_write(self.create_constraints)
        self.connection.execute_write(insert_data, self.create_first_node_query(), first_nodes)
        self.connection.execute_write(insert_data, self.create_first_relationships_query(), first_relationships)
        self.connection.execute_write(insert_data, self.create_second_node_query(), second_nodes)
        self.connection.execute_write(insert_data, self.create_second_relationships_query(), second_relationships)

    @abstractmethod
    def get_first_node_name(self):
        pass

    @abstractmethod
    def get_first_rels_name(self):
        pass

    @abstractmethod
    def get_second_node_name(self):
        pass

    @abstractmethod
    def get_second_rels_name(self):
        pass

    def create_constraints(self, tx):
        constraints = self.get_constraint_list()
        for constraint in constraints:
            tx.run(constraint)

    @abstractmethod
    def get_constraint_list(self):
        pass

    @abstractmethod
    def create_first_node_query(self):
        pass

    @abstractmethod
    def create_second_node_query(self):
        pass

    @abstractmethod
    def create_first_relationships_query(self):
        pass

    @abstractmethod
    def create_second_relationships_query(self):
        pass


class RoadBuildingsDbManager(TwoTypeNodeDBManager):
    def get_road_graph(self):
        g = ox.graph_from_place(self.city_name, simplify=True, retain_all=True, network_type="drive")

        gdf_nodes, gdf_relationships = ox.graph_to_gdfs(g)
        gdf_nodes.reset_index(inplace=True)
        gdf_relationships.reset_index(inplace=True)
        gdf_nodes["geometry_wkt"] = gdf_nodes["geometry"].apply(lambda x: x.wkt)
        gdf_relationships["geometry_wkt"] = gdf_relationships["geometry"].apply(lambda x: x.wkt)

        return gdf_nodes.drop(columns=["geometry"]), gdf_relationships.drop(columns=["geometry"])

    def get_graph(self):
        (road_nodes, road_relationships) = self.get_road_graph()
        tree = quads.QuadTree((0, 0), 200, 200)
        road_geo_to_node = {quads.Point(float(road_node.x), float(road_node.y)): road_node for road_node in
                            road_nodes.itertuples()}
        for road_node in road_nodes.itertuples():
            tree.insert((float(road_node.x), float(road_node.y)))
        buildings = pd.DataFrame(ox.geometries_from_place(self.city_name, tags={'building': True}))
        buildings = buildings.rename(columns={'addr:street': 'street', 'addr:housenumber': 'housenumber'})
        buildings_rels = []
        filtred_buildings = []
        i = 0
        for building in buildings.itertuples():
            print(i)
            i += 1
            if building.geometry.geom_type == "Polygon" or building.geometry.geom_type == "MultiPolygon":
                x = building.geometry.centroid.x
                y = building.geometry.centroid.y
            else:
                x, y = building.geometry.coords[0]
            nearest_point = tree.nearest_neighbors((x, y), 1)
            nearest_node = road_geo_to_node[nearest_point[0]]
            uniq_name = str(building.Index) + str(x) + str(y)
            nearest_node_filtred = {
                "intersect_osmid": nearest_node.osmid,
                "uniq_building_name": uniq_name,
                "uniq_building_road_name": uniq_name + "to" + str(nearest_node.x) + str(nearest_node.y),
                "length": math.sqrt((nearest_node.x - x) ** 2 + (nearest_node.y - y) ** 2)
            }
            buildings_rels.append(nearest_node_filtred)
            filtred_building = {
                "uniq_building_name": uniq_name,
                "building": getattr(building, "building", None),
                "name": getattr(building, "name", None),
                "street": getattr(building, "street", None),
                "housenumber": getattr(building, "housenumber", None),
                "amenity": getattr(building, "amenity", None),
                "phone": getattr(building, "phone", None),
                "shop": getattr(building, "shop", None),
                "year_of_construction": getattr(building, "year_of_construction", None),
                "opening_hours": getattr(building, "opening_hours", None),
                "x": x,
                "y": y
            }
            filtred_buildings.append(filtred_building)
        return (road_nodes, road_relationships, filtred_buildings, buildings_rels)

    def get_first_node_name(self):
        return self.city_name + "Intersection"

    def get_first_rels_name(self):
        return self.city_name + "Road"

    def get_second_node_name(self):
        return self.city_name + "Building"

    def get_second_rels_name(self):
        return self.city_name + "BuildingToNearestIntersection"

    def get_constraint_list(self):
        return [
            r"CREATE CONSTRAINT IF NOT EXISTS FOR (i:{}) REQUIRE i.osmid IS UNIQUE".format(self.db_graph_parameters.main_node_name),
            r"CREATE INDEX IF NOT EXISTS FOR ()-[r:{}]-() ON r.osmid".format(self.db_graph_parameters.main_rels_name),
            r"CREATE CONSTRAINT IF NOT EXISTS FOR (i:{}) REQUIRE i.uniq_building_name IS UNIQUE"
            .format(self.db_graph_parameters.secondary_node_name),
            r"CREATE INDEX IF NOT EXISTS FOR ()-[r:{}]-() ON r.uniq_building_road_name"
            .format(self.db_graph_parameters.secondary_rels_name),
        ]

    def create_first_node_query(self):
        return f'''
        UNWIND $rows AS row
        WITH row WHERE row.osmid IS NOT NULL
        MERGE (i:{self.db_graph_parameters.main_node_name} {{osmid: row.osmid}})
            SET i.location = point({{latitude: row.y, longitude: row.x }}),
                i.highway = row.highway,
                i.tram = row.tram,
                i.bus = row.bus,
                i.geometry_wkt = row.geometry_wkt,
                i.street_count = toInteger(row.street_count)
        RETURN COUNT(*) as total
        '''

    def create_second_node_query(self):
        return f"""
        UNWIND $rows AS row
        WITH row WHERE row.uniq_building_name IS NOT NULL
        MERGE (i:{self.db_graph_parameters.secondary_node_name} {{uniq_building_name: row.uniq_building_name}})
            SET i.building = row.building,
                i.name = row.name,
                i.street = row.street,
                i.housenumber = row.housenumber,
                i.amenity = row.amenity,
                i.phone = row.phone,
                i.shop = row.shop,
                i.year_of_construction = row.year_of_construction,
                i.opening_hours = row.opening_hours,
                i.location = point({{latitude: row.y, longitude: row.x }})
        RETURN COUNT(*) as total
    """

    def create_first_relationships_query(self):
        return f'''
        UNWIND $rows AS path
        MATCH (u:{self.db_graph_parameters.main_node_name} {{osmid: path.u}})
        MATCH (v:{self.db_graph_parameters.main_node_name} {{osmid: path.v}})
        MERGE (u)-[r:{self.db_graph_parameters.main_rels_name} {{osmid: path.osmid}}]->(v)
            SET r.name = path.name,
                r.highway = path.highway,
                r.railway = path.railway,
                r.oneway = path.oneway,
                r.lanes = path.lanes,
                r.max_speed = path.maxspeed,
                r.geometry_wkt = path.geometry_wkt,
                r.length = toFloat(path.length)
        RETURN COUNT(*) AS total
        '''

    def create_second_relationships_query(self):
        return f'''
        UNWIND $rows AS path
        MATCH (u:{self.db_graph_parameters.main_node_name} {{osmid: path.intersect_osmid}})
        MATCH (v:{self.db_graph_parameters.secondary_node_name} {{ uniq_building_name: path.uniq_building_name}})
        MERGE (u)-[r:{self.db_graph_parameters.secondary_rels_name} {{uniq_building_road_name: path.uniq_building_road_name}}]->(v)
            SET r.intersect_osmid = path.intersect_osmid,
                r.uniq_building_name = path.uniq_building_name,
                r.length = toFloat(path.length)
        RETURN COUNT(*) AS total
    '''

    def get_main_node_name(self):
        return self.get_main_node_name().replace('-', '')

    def get_main_rels_name(self):
        return self.get_main_rels_name().replace('-', '')

    def get_weight(self):
        return "length"


class OneTypeNodeDBManager(GraphDBManager):
    def __init__(self, graph_analisis_context: GraphAnalysContext):
        super().__init__(graph_analisis_context)

    def update_db(self, city_name):
        (nodes, relationships) = self.get_graph()
        if nodes is None and relationships is None:
            print("Graph for", city_name, "is empty!")
            return
        self.connection.execute_write(self.create_constraints)
        self.connection.execute_write(insert_data, self.create_node_query(), nodes)
        self.connection.execute_write(insert_data, self.create_relationships_query(), relationships)

    @abstractmethod
    def get_bd_all_node_query_graph(self):
        pass

    @abstractmethod
    def get_bd_all_rels_query_graph(self):
        pass

    @abstractmethod
    def get_node_name(self):
        pass

    @abstractmethod
    def get_rels_name(self):
        pass

    def get_bd_all_node_graph(self):
        node_get_query = self.get_bd_all_node_query_graph()
        return self.connection.read_all(node_get_query)

    def get_bd_all_rels_graph(self):
        rels_get_query = self.get_bd_all_rels_query_graph()
        return self.connection.read_all(rels_get_query)

    def create_constraints(self, tx):
        constraints = self.get_constraint_list()
        for constraint in constraints:
            tx.run(constraint)

    def get_main_node_name(self):
        return self.get_node_name().replace('-', '')

    def get_main_rels_name(self):
        return self.get_rels_name().replace('-', '')

    @abstractmethod
    def get_constraint_list(self):
        pass

    @abstractmethod
    def create_node_query(self):
        pass

    @abstractmethod
    def create_relationships_query(self):
        pass


class TransportNetworkGraphDBManager(OneTypeNodeDBManager):
    def create_node_query(self):
        return f'''
            UNWIND $rows AS row
            WITH row WHERE row.name IS NOT NULL
            MERGE (s:{self.db_graph_parameters.main_node_name} {{name: row.name}})
                SET s.location = point({{latitude: row.yCoordinate, longitude: row.xCoordinate }}),
                    s.roteList = row.roteList,
                    s.isCoordinateApproximate = row.isCoordinateApproximate
            RETURN COUNT(*) AS total
        '''

    def create_relationships_query(self):
        return f'''
            UNWIND $rows AS path
            MATCH (u:{self.db_graph_parameters.main_node_name} {{name: path.startStop}})
            MATCH (v:{self.db_graph_parameters.main_node_name} {{name: path.endStop}})
            MERGE (u)-[r:{self.db_graph_parameters.main_rels_name} {{name: path.name}}]->(v)
                SET r.duration = path.duration,
                    r.route = path.route
            RETURN COUNT(*) AS total
        '''

    def get_bd_all_node_query_graph(self):
        return f'''
        MATCH (s:Stop)
        RETURN 
            ID(s) AS id,
            s.roteList AS roteList, 
            s.location.longitude AS x, 
            s.location.latitude AS y, 
            s.name AS name, 
            s.isCoordinateApproximate AS isCoordinateApproximate
        '''

    def get_bd_all_rels_query_graph(self):
        return f'''
        MATCH (u:Stop)-[r:{self.db_graph_parameters.main_rels_name}]->(v:Stop) 
        RETURN
            u.name AS first_stop_name, 
            v.name AS second_stop_name, 
            r.duration AS duration
        '''

    def get_constraint_list(self):
        return [
            f"CREATE CONSTRAINT IF NOT EXISTS FOR (s:{self.db_graph_parameters.main_node_name}) REQUIRE s.name IS UNIQUE",
            f"CREATE INDEX IF NOT EXISTS FOR ()-[r:{self.db_graph_parameters.main_rels_name}]-() ON r.name"
        ]

    @abstractmethod
    def get_graph(self):
        pass

    @abstractmethod
    def get_node_name(self):
        pass

    @abstractmethod
    def get_rels_name(self):
        pass

    @abstractmethod
    def get_weight(self):
        pass


class RoadGraphDBManager(OneTypeNodeDBManager):

    def get_graph(self):
        g = ox.graph_from_place(self.city_name, simplify=True, retain_all=True, network_type="drive")

        gdf_nodes, gdf_relationships = ox.graph_to_gdfs(g)
        gdf_nodes.reset_index(inplace=True)
        gdf_relationships.reset_index(inplace=True)
        gdf_nodes["geometry_wkt"] = gdf_nodes["geometry"].apply(lambda x: x.wkt)
        gdf_relationships["geometry_wkt"] = gdf_relationships["geometry"].apply(lambda x: x.wkt)

        return gdf_nodes.drop(columns=["geometry"]), gdf_relationships.drop(columns=["geometry"])

    def create_node_query(self):
        return f'''
        UNWIND $rows AS row
        WITH row WHERE row.osmid IS NOT NULL
        MERGE (i:{self.db_graph_parameters.main_node_name} {{osmid: row.osmid}})
            SET i.location = point({{latitude: row.y, longitude: row.x }}),
                i.highway = row.highway,
                i.tram = row.tram,
                i.bus = row.bus,
                i.geometry_wkt = row.geometry_wkt,
                i.street_count = toInteger(row.street_count)
        RETURN COUNT(*) as total
        '''

    def create_relationships_query(self):
        return f'''
        UNWIND $rows AS path
        MATCH (u:{self.db_graph_parameters.main_node_name} {{osmid: path.u}})
        MATCH (v:{self.db_graph_parameters.main_node_name} {{osmid: path.v}})
        MERGE (u)-[r:{self.db_graph_parameters.main_rels_name} {{osmid: path.osmid}}]->(v)
            SET r.name = path.name,
                r.highway = path.highway,
                r.railway = path.railway,
                r.oneway = path.oneway,
                r.lanes = path.lanes,
                r.max_speed = path.maxspeed,
                r.geometry_wkt = path.geometry_wkt,
                r.length = toFloat(path.length)
        RETURN COUNT(*) AS total
        '''

    def get_constraint_list(self):
        return [
            f"CREATE CONSTRAINT IF NOT EXISTS FOR (i:{self.db_graph_parameters.main_node_name}) REQUIRE i.osmid IS UNIQUE",
            f"CREATE INDEX IF NOT EXISTS FOR ()-[r:{self.db_graph_parameters.main_rels_name}]-() ON r.osmid"
        ]

    def get_bd_all_node_query_graph(self):
        return f'''
        MATCH (s:{self.db_graph_parameters.main_node_name})
        RETURN 
            ID(s) AS id,
            s.highway AS highway,
            s.location.longitude AS x, 
            s.location.latitude AS y, 
            s.tram AS tram,
            s.bus AS bus, 
            s.geometry_wkt AS geometry_wkt,
            s.street_count AS street_count
        '''

    def get_bd_all_rels_query_graph(self):
        return f'''
        MATCH (u:{self.db_graph_parameters.main_node_name})-[r:{self.db_graph_parameters.main_rels_name}]->(v:{self.db_graph_parameters.main_node_name}) 
        RETURN
            u.osmid AS first_osmid, 
            v.osmid AS second_osmid, 
            r.name AS name,
            r.highway AS highway,
            r.railway AS railway,
            r.oneway AS oneway,
            r.lanes AS lanes,
            r.max_speed AS maxspeed,
            r.geometry_wkt AS geometry_wkt,
            r.length AS length
        '''

    def get_node_name(self):
        return self.city_name + "Intersection"

    def get_rels_name(self):
        return self.city_name + "RoadSegment"

    def get_weight(self):
        return "length"

    def node_geometry_identity(self):
        return "geometry_wkt"


class BusGraphDBManager(TransportNetworkGraphDBManager):

    def get_graph(self):
        parser = BusGraphParser(self.city_name)
        (nodes, relationships) = parser.parse()
        return list(nodes.values()), relationships

    def get_node_name(self):
        return self.city_name + "BusStop"

    def get_rels_name(self):
        return self.city_name + "BusRouteSegment"

    def get_weight(self):
        return "duration"

    def node_geometry_identity(self):
        return "location"


class TrolleyGraphDBManager(TransportNetworkGraphDBManager):
    def get_graph(self):
        parser = TrolleyGraphParser(self.city_name)
        (nodes, relationships) = parser.parse()
        return list(nodes.values()), relationships

    def get_node_name(self):
        return self.city_name + "TrolleyStop"

    def get_rels_name(self):
        return self.city_name + "TrolleyRouteSegment"

    def get_weight(self):
        return "duration"


class TramGraphDBManager(TransportNetworkGraphDBManager):
    def get_graph(self):
        parser = TramGraphParser(self.city_name)
        (nodes, relationships) = parser.parse()
        return list(nodes.values()), relationships

    def get_node_name(self):
        return self.city_name + "TramStop"

    def get_rels_name(self):
        return self.city_name + "TramRouteSegment"

    def get_weight(self):
        return "duration"


class MiniBusGraphDBManager(TransportNetworkGraphDBManager):
    def get_graph(self):
        parser = MiniBusGraphParser(self.city_name)
        (nodes, relationships) = parser.parse()
        return list(nodes.values()), relationships

    def get_node_name(self):
        return self.city_name + "MiniBusStop"

    def get_rels_name(self):
        return self.city_name + "MiniBusRouteSegment"

    def get_weight(self):
        return "duration"


def insert_data(tx, query, rows, batch_size=10000):
    total = 0
    batch = 0

    df = pd.DataFrame(rows)

    while batch * batch_size < len(df):
        current_batch = df.iloc[batch * batch_size:(batch + 1) * batch_size]
        batch_data = current_batch.to_dict('records')
        results = tx.run(query, parameters={'rows': batch_data}).data()
        print(results)
        total += results[0]['total']
        batch += 1
    return total



from urban_transit_network_analysis.enums.ClusteringTypes import ClusteringTypes
from urban_transit_network_analysis.enums.HeatMapMetrics import HeatMapMetrics
from urban_transit_network_analysis.enums.HistogramMetrics import HistogramMetrics

"""
    Контекст для отрисовки графиков о метриках сетей
"""


class PrintGraphAnalysContext:
    def __init__(
            self,
            heat_map_metrics_list: [HeatMapMetrics] = None,
            histogram_map_metrics_list: [HistogramMetrics] = None,
            map_metric_list: [HeatMapMetrics] = None,
            map_clustering_list: [ClusteringTypes] = None,
            need_create_map_without_metrics: bool = None,
            mesh_size: int = 100
    ):
        self.heat_map_metrics_list = heat_map_metrics_list if heat_map_metrics_list is not None \
            else list(HeatMapMetrics.__members__.values())
        self.histogram_map_metrics_list = histogram_map_metrics_list if histogram_map_metrics_list is not None \
            else list(HistogramMetrics.__members__.values())
        self.map_metric_list = map_metric_list if map_metric_list is not None \
            else list(HeatMapMetrics.__members__.values())
        self.map_clustering_list = map_clustering_list if map_clustering_list is not None \
            else list(ClusteringTypes.__members__.values())
        self.need_create_map_without_metrics = need_create_map_without_metrics
        self.mesh_size = mesh_size



import datetime
import json
import math
import os
import re
import time
from abc import abstractmethod

import requests
from bs4 import BeautifulSoup
"""
    Класс занимающийся парсингом данных с сайта https://kudikina.ru
"""

cache_file = '../cache/city_urls.json'
cache_expire_days = 30
site_url = "https://kudikina.ru"
map_url = "/map"
timetable_forward_url = "/A"
timetable_backward_url = "/B"
# TODO: need to update according to city coordinates
city_avg_x_coordinate = 60.0
city_avg_y_coordinate = 30.0
request_pause_sec = 2


class AbstractTransportGraphParser:

    def __init__(self, city_name):
        self.city_name = city_name
        self.city_url = self.__get_city_url()
        self.nodes = {}
        self.relationships = []
        self.transport_url = self.get_transport_url()
        self.transport_class = self.get_transport_class()

    def parse(self):
        if self.city_url is None:
            return None, None

        for route_info in self.get_all_routes_info():
            route_name = route_info[0]
            route_url = route_info[2]

            self.__add_stops_and_routes(route_name, route_url)

            print(route_url)
            time.sleep(request_pause_sec)

        return self.nodes, self.relationships

    def __get_city_url(self):
        cities_url = self.load_cache(cache_file)
        if not cities_url:
            print("Cities url cache is expired or empty, lets fill it.")
            cities_url = self.parse_all_city_urls()
            self.save_cache(cache_file, cities_url)
            print("Cities url are saved in cache.")
        city_url = cities_url.get(self.city_name)
        if city_url is None:
            print('No such city in parsed data')
        return city_url

    def __add_stops_and_routes(self, route_name, route_url):
        (timetable, successes_parse) = self.get_timetable(route_url)
        if successes_parse is False:
            return

        stop_coordinates = self.get_stop_coordinates(route_url)
        last_coordinate = Coordinate(city_avg_x_coordinate, city_avg_y_coordinate)
        previous_transport_stop_name = None
        previous_time_point = None

        for row in timetable:
            transport_stop_name = row["stopName"]
            time_point = row["timePoint"]
            coordinate = self.__get_filled_coordinate(stop_coordinates, transport_stop_name, last_coordinate)

            transport_stop = self.__update_or_add_stop(transport_stop_name, coordinate, route_name)

            if previous_transport_stop_name is not None:
                duration = self.calculate_duration(previous_time_point, time_point)
                if duration is False:
                    continue
                self.__add_route(
                    previous_transport_stop_name,
                    transport_stop,
                    duration,
                    route_name
                )

            last_coordinate = coordinate
            previous_transport_stop_name = transport_stop
            previous_time_point = time_point

    def __get_filled_coordinate(self, stop_coordinates, stop_name, last_coordinate):
        coordinate = stop_coordinates.get(stop_name)
        if coordinate is None or not coordinate.is_defined():
            coordinate = Coordinate(last_coordinate.x, last_coordinate.y, True)
        return coordinate

    def __update_or_add_stop(self, transport_stop_name, coordinate, route_name):

        transport_stop_name, is_new_stop = self.__check_and_find_unique_stop(transport_stop_name, coordinate)

        if not is_new_stop:
            transport_stop = self.nodes.get(transport_stop_name)
            transport_stop["roteList"].append(route_name)
        else:
            transport_stop = {
                "name": transport_stop_name,
                "roteList": [route_name],
                "xCoordinate": coordinate.x,
                "yCoordinate": coordinate.y,
                "isCoordinateApproximate": coordinate.is_approximate
            }
            self.nodes[transport_stop_name] = transport_stop
        return transport_stop

    def __check_and_find_unique_stop(self, transport_stop_name, coordinate):
        is_new_stop = True
        while self.nodes.get(transport_stop_name) is not None:
            transport_stop = self.nodes[transport_stop_name]
            old_coordinate = Coordinate(transport_stop["xCoordinate"], transport_stop["yCoordinate"])
            if self.are_stops_same(old_coordinate, coordinate):
                is_new_stop = False
                break
            else:
                transport_stop_name = self.increment_suffix(transport_stop_name)

        return transport_stop_name, is_new_stop

    def __add_route(self, start_stop, end_stop, duration, route_name):
        if start_stop is not None and end_stop is not None:
            self.relationships.append({"startStop": start_stop["name"],
                                       "endStop": end_stop["name"],
                                       "name": start_stop["name"] + " -> " + end_stop[
                                           "name"] + "; route_name: " + route_name,
                                       "route": route_name,
                                       "duration": duration
                                       })

    def get_all_routes_info(self):
        if self.city_url is None:
            return []

        full_url = site_url + self.city_url + self.transport_url

        response = requests.get(full_url)
        html = response.text
        soup = BeautifulSoup(html, "html.parser")

        transport_list = []

        bus_items = soup.find_all("a", class_=self.transport_class)
        for item in bus_items:
            transport_number = item.text.strip()
            transport_route = item.find("span").text.strip()
            href_link = item["href"]
            transport_list.append([transport_number, transport_route, href_link])
        return transport_list

    def get_timetable(self, route_url):
        (timetable1, successes_parse1) = self.get_one_direction_timetable(route_url, timetable_forward_url)
        (timetable2, successes_parse2) = self.get_one_direction_timetable(route_url, timetable_backward_url)
        if successes_parse1 and successes_parse2:
            return timetable1 + timetable2, True
        else:
            return None, False

    def get_one_direction_timetable(self, route_url, timetable_url):
        full_url = site_url + route_url + timetable_url

        response = requests.get(full_url)
        soup = BeautifulSoup(response.text, 'html.parser')

        stop_times = []
        for stop_div in soup.find_all('div', class_='bus-stop'):
            name = stop_div.find('a').text.strip()
            time_point = stop_div.find_next_sibling('div', class_='col-xs-12').find('span')
            if time_point is not None:
                parsed_time_point = time_point.text.strip()
                if parsed_time_point[len(parsed_time_point) - 1] == 'K':
                    parsed_time_point = parsed_time_point[:-1]
            else:
                return None, False
            clean_name = re.sub(r"\d+\) ", "", name)
            stop_times.append({"stopName": clean_name, "timePoint": parsed_time_point})
        return stop_times, True

    def get_stop_coordinates(self, route_url):
        full_url = site_url + route_url + map_url
        response = requests.get(full_url)
        soup = BeautifulSoup(response.text, 'html.parser')

        script_tags = soup.find_all('script', type="text/javascript")
        script_tag = None

        for tag in script_tags:
            if 'drawMap' in tag.text:
                script_tag = tag
                break

        if script_tag:
            script_text = script_tag.text
            coordinates = self.extract_coordinates(script_text)

            return coordinates
        else:
            return {}

    def extract_coordinates(self, script_text):
        matches = re.findall(r'{"name":\s*"(.*?)",\s*"lat":\s*(-?\d+\.?\d*),?\s*"long":\s*(-?\d+\.?\d*)?}', script_text)

        coordinates = {}
        for match in matches:
            name = match[0].replace("\\", "")
            # `match` contains latitude and longitude which equals to y and x coordinates
            x = float(match[2])
            y = float(match[1])
            coordinates[name] = Coordinate(x, y)

        return coordinates

    def load_cache(self, cache_file):
        if os.path.exists(cache_file):
            modification_time = os.path.getmtime(cache_file)
            current_time = datetime.datetime.now()
            if (current_time - datetime.datetime.fromtimestamp(modification_time)).days <= cache_expire_days:
                with open(cache_file, 'r') as file:
                    return json.load(file)
        return {}

    def save_cache(self, cache_file, cache_data):
        cache_dir = os.path.dirname(cache_file)
        if cache_dir and not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        with open(cache_file, 'w') as file:
            json.dump(cache_data, file)

    def parse_all_city_urls(self):
        url = "https://kudikina.ru/"
        response = requests.get(url)
        time.sleep(2)
        html_content = response.text
        soup = BeautifulSoup(html_content, 'html.parser')
        cities = {}

        for li in soup.find_all('ul', class_='list-unstyled cities block-regions'):
            for region in li.find_all('a'):
                region_name = region.find('span', class_='city-name').text.strip()
                region_href = region['href']
                region_response = requests.get(url[:-1] + region_href)
                region_html_content = region_response.text
                region_soup = BeautifulSoup(region_html_content, 'html.parser')
                city_list = region_soup.find_all('ul', class_='list-unstyled cities')
                time.sleep(2)
                if len(city_list) == 0:
                    cities[region_name] = region_href
                    print(region_href + ' Was parsed')
                else:
                    region_cities = city_list[0].find_all('a')
                    for city in region_cities:
                        city_name = city.find('span', class_='city-name').text.strip()
                        city_href = city['href']
                        cities[city_name] = city_href
                        print(city_href + ' Was parsed')
        return cities

    def calculate_duration(self, start_stop, end_stop):
        try:
            start_hour, start_minute = map(int, start_stop.split(':'))
            end_hour, end_minute = map(int, end_stop.split(':'))
            return abs((end_hour * 60 + end_minute) - (start_hour * 60 + start_minute))
        except ValueError:
            return False
        except AttributeError:
            return False
    def are_stops_same(self, coord1, coord2, tolerance=0.005):
        distance = math.dist(coord1.get_xy(), coord2.get_xy())
        return abs(distance) < tolerance

    def increment_suffix(self, name):
        if name and name[-1].isdigit():
            index = len(name) - 1
            while index >= 0 and name[index].isdigit():
                index -= 1
            number = int(name[index + 1:]) + 1
            return f"{name[:index + 1]}{number}"
        else:
            return f"{name} 1"

    @abstractmethod
    def get_transport_class(self):
        pass

    @abstractmethod
    def get_transport_url(self):
        pass


class BusGraphParser(AbstractTransportGraphParser):
    def get_transport_class(self):
        return "bus-item bus-icon"

    def get_transport_url(self):
        return "bus/"


class TrolleyGraphParser(AbstractTransportGraphParser):
    def get_transport_url(self):
        return "trolley/"

    def get_transport_class(self):
        return "bus-item trolley-icon"


class BusGraphParser(AbstractTransportGraphParser):

    def get_transport_url(self):
        return "bus/"

    def get_transport_class(self):
        return "bus-item bus-icon"


class MiniBusGraphParser(AbstractTransportGraphParser):
    def get_transport_url(self):
        return "mtaxi/"

    def get_transport_class(self):
        return "bus-item mtaxi-icon"


class TramGraphParser(AbstractTransportGraphParser):
    def get_transport_url(self):
        return "tram/"

    def get_transport_class(self):
        return "bus-item tram-icon"


class Coordinate:
    def __init__(self, x=None, y=None, is_approximate=False):
        self.x = x
        self.y = y
        self.is_approximate = is_approximate

    def __str__(self):
        return f"({self.x}, {self.y})"

    def is_defined(self):
        if self.x is None or self.y is None:
            return False
        else:
            return True

    def get_xy(self):
        return [self.x, self.y]



def to_bus_graph(nodes, rels):
    rels_list = {}
    for node in nodes.values:
        rels_list[node[4]] = []
    for rel in rels.values:
        rels_list[rel[0]].append(rel[1])
    return nodes, rels_list



from urban_transit_network_analysis.database.Neo4jConnection import Neo4jConnection
"""
    Класс содержащий query для вычисления кластеризации сети 
"""


class CommunityDetection:
    def __init__(self, algorithm_name, property_name):
        self.algorithm_name = algorithm_name
        self.property_name = property_name
        self.connection = Neo4jConnection()

    def detect_communities(
            self,
            graph_name,
            relationship_weight_property
    ):
        self.__detect_communities(graph_name, relationship_weight_property)
        return self.__write_communities(graph_name)

    def __detect_communities(self, graph_name, relationship_weight_property):
        query = f'''
            CALL gds.{self.algorithm_name}.stream(
                '{graph_name}',
                {{
                    relationshipWeightProperty: '{relationship_weight_property}'
                }}
            )
            YIELD nodeId, communityId
            RETURN communityId, COUNT(DISTINCT nodeId) AS members
            ORDER BY members DESC
        '''
        return self.connection.run(query)

    def __write_communities(self, graph_name):
        query = f'''
            CALL gds.{self.algorithm_name}.write(
                '{graph_name}', 
                {{
                    writeProperty: '{self.property_name}'
                }}
            ) 
            YIELD communityCount, modularity, modularities
        '''
        return self.connection.execute_query(query).records[0][2]


class Leiden(CommunityDetection):
    def __init__(self):
        super().__init__("leiden", "leiden_community")


class Louvain(CommunityDetection):
    def __init__(self):
        super().__init__("louvain", "louvain_community")



from enum import Enum
import urban_transit_network_analysis.database.GraphDbManager as GraphDbManager
"""
    Список типов сетей с соответсвующими конструкторами для классов работающими с бд
"""


class GraphTypes(Enum):
    ROAD_GRAPH = GraphDbManager.RoadGraphDBManager
    ROAD_BUILDINGS_GRAPH = GraphDbManager.RoadBuildingsDbManager
    BUS_GRAPH = GraphDbManager.BusGraphDBManager
    TRAM_GRAPH = GraphDbManager.TramGraphDBManager
    TROLLEY_GRAPH = GraphDbManager.TrolleyGraphDBManager
    MINIBUS_GRAPH = GraphDbManager.MiniBusGraphDBManager



from urban_transit_network_analysis.database.Neo4jConnection import Neo4jConnection

from urban_transit_network_analysis.context.GraphAnalysContext import GraphAnalysContext


class AnalysPreparer:
    def __init__(self, graph_analisis_context: GraphAnalysContext):
        self.connection = Neo4jConnection()
        self.graph_db_parameters = graph_analisis_context.neo4j_DB_graph_parameters
        self.graph_name = graph_analisis_context.graph_name

    def prepare(self):
        self.make_graph()

    def make_graph(self):
        query = f'''
            CALL gds.graph.project(
            '{self.graph_name}',
            '{self.graph_db_parameters.main_node_name}',
            {{
                {self.graph_db_parameters.main_rels_name}: {{
                    orientation: 'UNDIRECTED',
                    properties: '{self.graph_db_parameters.weight}'
                }}
            }}
        )
        '''
        self.connection.run(query)


from abc import abstractmethod

from urban_transit_network_analysis.context import DBGraphParameters
from urban_transit_network_analysis.database.Neo4jConnection import Neo4jConnection
"""
    Класс содержащий query для вычисления распределения метрик сети 
"""



class MetricsDistributionNode:
    def __init__(self, db_params: DBGraphParameters):
        self.node_name = db_params.main_node_name
        self.rels_name = db_params.main_rels_name
        self.node_geometry_identity = db_params.node_geometry_identity
        self.metrics_calculate = self.metrics_calculate()
        self.connection = Neo4jConnection()

    @abstractmethod
    def metrics_calculate(self):
        pass


    def calculate_distribution(self, needLog = False):
        query = f'''
            MATCH (first_node:{self.node_name})
            WITH first_node, {self.metrics_calculate} AS Metric
            RETURN first_node.{self.node_geometry_identity} AS NodeIdentity, Metric
            ORDER BY Metric DESC
        '''
        return self.connection.execute_query(query, needLog).records

class DegreeDistribution(MetricsDistributionNode):
    def metrics_calculate(self):
        return 'count(rels)'

    def calculate_distribution(self, needLog = False):
        query = f'''
            MATCH (first_node:{self.node_name})-[rels:{self.rels_name}]-(second_node:{self.node_name})
            WITH first_node, {self.metrics_calculate} AS Metric
            RETURN first_node.{self.node_geometry_identity} AS NodeIdentity, Metric
            ORDER BY Metric DESC
        '''
        return self.connection.execute_query(query, needLog).records

class PageRankDistribution(MetricsDistributionNode):
    def metrics_calculate(self):
        return 'first_node.pageRank'

class BetweennessDistribution(MetricsDistributionNode):
    def metrics_calculate(self):
        return 'first_node.betweenness'

class LouvainClusteringDistribution(MetricsDistributionNode):
    def metrics_calculate(self):
        return 'first_node.louvain_community'

class LeidenClusteringDistribution(MetricsDistributionNode):
    def metrics_calculate(self):
        return 'first_node.leiden_community'



import random

from urban_transit_network_analysis.context.DBGraphParameters import DBGraphParameters
from urban_transit_network_analysis.context.MetricCalculationContext import MetricCalculationContext
from urban_transit_network_analysis.context.PrintGraphAnalysContext import PrintGraphAnalysContext
from urban_transit_network_analysis.enums.GraphTypes import GraphTypes

"""
    Контекст для анализа сети конкретной сети
"""


class GraphAnalysContext:
    def __init__(
            self,
            metric_calculation_context: MetricCalculationContext = MetricCalculationContext(),
            print_graph_analis_context: PrintGraphAnalysContext = PrintGraphAnalysContext(),
            graph_name: str = "SomeGraph" + str(random.random()),
            graph_type: GraphTypes = GraphTypes.ROAD_GRAPH,
            need_prepare_data: bool = True,
            need_calculate_and_print_data: bool = True,
            need_create_graph: bool = True,
            city_name: str = ""
    ):
        self.metric_calculation_context = metric_calculation_context
        self.graph_name = graph_name
        self.graph_type = graph_type
        self.need_prepare_data = need_prepare_data
        self.print_graph_analis_context = print_graph_analis_context
        self.need_calculate_and_print_data = need_calculate_and_print_data
        self.need_create_graph = need_create_graph
        self.neo4j_DB_graph_parameters = DBGraphParameters()
        self.city_name = city_name



"""
    Контекст для вычисления метрик сетей
"""


class MetricCalculationContext:
    def __init__(
            self,
            need_leiden_community_id: bool = True,
            need_louvain_community_id: bool = True,
            need_leiden_modulariry: bool = True,
            need_louvain_modulariry: bool = True,
            need_betweenessens: bool = True,
            need_page_rank: bool = True,
            need_degree: bool = True
    ):
        self.need_leiden_community_id = need_leiden_community_id
        self.need_louvain_community_id = need_louvain_community_id
        self.need_leiden_modulariry = need_leiden_modulariry
        self.need_louvain_modulariry = need_louvain_modulariry
        self.need_betweenessens = need_betweenessens
        self.need_page_rank = need_page_rank
        self.need_degree = need_degree



from urban_transit_network_analysis.context import *
from urban_transit_network_analysis.data import *
from urban_transit_network_analysis.database import *
from urban_transit_network_analysis.enums import *
from urban_transit_network_analysis.graphics import *
from urban_transit_network_analysis.parser import *
from urban_transit_network_analysis.utils import *



from urban_transit_network_analysis.context import GraphAnalysContext
from urban_transit_network_analysis.database.CommunityDetection import Leiden, Louvain
from urban_transit_network_analysis.database.MetricsCalculate import Betweenness, PageRank
"""
    Класс записывающий метрики сетей в бд
"""


class MetricDataPreparer:
    def __init__(self, graph_analisis_context: GraphAnalysContext):
        self.leiden_calculator = None
        self.louvain_calculator = None
        self.betweenessens_calculator = None
        self.page_rank_calculator = None
        metric_calculation_context = graph_analisis_context.metric_calculation_context
        if metric_calculation_context.need_leiden_community_id or metric_calculation_context.need_leiden_modulariry:
            self.leiden_calculator = Leiden()
        if metric_calculation_context.need_louvain_community_id or metric_calculation_context.need_louvain_modulariry:
            self.louvain_calculator = Louvain()
        if metric_calculation_context.need_betweenessens:
            self.betweenessens_calculator = Betweenness()
        if metric_calculation_context.need_page_rank:
            self.page_rank_calculator = PageRank()
        self.graph_analisis_context = graph_analisis_context

    def prepare_metrics(self):
        result = {}
        if self.leiden_calculator is not None:
            result["leiden_modularity_value"] = self.prepare_leiden()
        if self.louvain_calculator is not None:
            result["louvain_modularity_value"] = self.prepare_louvain()
        if self.betweenessens_calculator is not None:
            self.prepare_betweenessens()
        if self.page_rank_calculator is not None:
            self.prepare_page_rank()
        return result

    def prepare_leiden(self):
        result = self.leiden_calculator.detect_communities(
            self.graph_analisis_context.graph_name,
            self.graph_analisis_context.neo4j_DB_graph_parameters.weight
        )
        print(result)
        print(f"LeidenAlgorithm Community detection for graph {self.graph_analisis_context.graph_name} completed.")
        return result

    def prepare_louvain(self):
        result = self.louvain_calculator.detect_communities(
            self.graph_analisis_context.graph_name,
            self.graph_analisis_context.neo4j_DB_graph_parameters.weight
        )
        print(result)
        print(f"LovainAlgorithm Community detection for graph {self.graph_analisis_context.graph_name} completed.")
        return result

    def prepare_betweenessens(self):
        self.betweenessens_calculator.metric_calculate(
            self.graph_analisis_context.graph_name,
            self.graph_analisis_context.neo4j_DB_graph_parameters.weight
        )
        print(f"betweenessens metric calculated for graph {self.graph_analisis_context.graph_name}.")

    def prepare_page_rank(self):
        self.page_rank_calculator.metric_calculate(
            self.graph_analisis_context.graph_name,
            self.graph_analisis_context.neo4j_DB_graph_parameters.weight
        )
        print(f"pageRank metric calculated for graph {self.graph_analisis_context.graph_name}.")




from enum import Enum

"""
    Список метрик с названиями для отрисовки в Histogram
"""


class HistogramMetrics(Enum):
    PAGE_RANK = 'page_rank'
    BEETWEENESSENS = 'beetweenessens'
    DEGREE = 'degree'
    LEIDEN_MODULARITY = 'leiden_modularity'
    LOUVAIN_MODULARITY = 'louvain_modularity'



from urban_transit_network_analysis.context.MetricCalculationContext import MetricCalculationContext
from urban_transit_network_analysis.context.GraphAnalysContext import GraphAnalysContext

"""
    Контекст для анализа сетей
"""


class AnalysContext:
    def __init__(
            self,
            ru_city_name: str = None,
            common_metric_calculation_context: MetricCalculationContext = MetricCalculationContext(),
            graph_analis_context: [GraphAnalysContext] = None
    ):
        if graph_analis_context is None:
            graph_analis_context = [GraphAnalysContext()]
        self.ru_city_name = ru_city_name
        self.common_metric_calculation_context = common_metric_calculation_context
        for item in graph_analis_context:
            if item.metric_calculation_context is None:
                item.metric_calculation_context = common_metric_calculation_context
        self.graph_analis_context = graph_analis_context



from urban_transit_network_analysis.context import GraphAnalysContext
from urban_transit_network_analysis.database.MetricsDistribution import DegreeDistribution, BetweennessDistribution, PageRankDistribution, \
    LeidenClusteringDistribution, LouvainClusteringDistribution

"""
    Класс вычисляющий метрики сетей(берёт уже записанные метрики из бд или вычисляет не сложные)
"""


class MetricDataCalculator:
    def __init__(
            self,
            graph_analisis_context: GraphAnalysContext
    ):
        metric_calculation_context = graph_analisis_context.metric_calculation_context
        self.degree_distibution_calculator = None
        self.betweenessens_distribution_calculator = None
        self.page_rank_distribution_calculator = None
        db_parameters = graph_analisis_context.neo4j_DB_graph_parameters
        if metric_calculation_context.need_degree:
            self.degree_distibution_calculator = DegreeDistribution(db_parameters)
        if metric_calculation_context.need_betweenessens:
            self.betweenessens_distribution_calculator = BetweennessDistribution(db_parameters)
        if metric_calculation_context.need_page_rank:
            self.page_rank_distribution_calculator = PageRankDistribution(db_parameters)
        if metric_calculation_context.need_leiden_community_id:
            self.leiden_community_id_calculator = LeidenClusteringDistribution(db_parameters)
        if metric_calculation_context.need_louvain_community_id:
            self.louvain_community_id_calculator = LouvainClusteringDistribution(db_parameters)

    def calculate_data(self, prepare_result: dict):
        if prepare_result is None:
            prepare_result = {}
        degree_distribution = {}
        if self.degree_distibution_calculator is not None:
            degree_distirbution_data = self.degree_distibution_calculator.calculate_distribution()
            degree_distribution = {"degree_value": [item[1] for item in degree_distirbution_data] }

        betweenessens_distibution = {}
        if self.betweenessens_distribution_calculator is not None:
            betweenessens_distirbution_data = self.betweenessens_distribution_calculator.calculate_distribution()
            betweenessens_distibution = {
                "beetweenessens_identity": [convert_to_point(item[0]) for item in betweenessens_distirbution_data],
                "beetweenessens_value": [item[1] for item in betweenessens_distirbution_data],
            }

        page_rank_distibution = {}
        if self.page_rank_distribution_calculator is not None:
            page_rank_distirbution_data = self.page_rank_distribution_calculator.calculate_distribution()
            page_rank_distibution = {
                "page_rank_identity": [convert_to_point(item[0]) for item in page_rank_distirbution_data],
                "page_rank_value": [item[1] for item in page_rank_distirbution_data],
            }

        leiden_distirbution = {}
        if self.leiden_community_id_calculator is not None:
            leiden_distirbution_data = self.leiden_community_id_calculator.calculate_distribution()
            leiden_distirbution = {
                "leiden_identity": [convert_to_point(item[0]) for item in leiden_distirbution_data],
                "leiden_value": [item[1] for item in leiden_distirbution_data],
            }

        louvain_distirbution = {}
        if self.louvain_community_id_calculator is not None:
            louvain_distirbution = self.louvain_community_id_calculator.calculate_distribution()
            louvain_distirbution = {
                "louvain_identity": [convert_to_point(item[0]) for item in louvain_distirbution],
                "louvain_value": [item[1] for item in louvain_distirbution],
            }
        return {
            **degree_distribution,
            **betweenessens_distibution,
            **page_rank_distibution,
            **prepare_result,
            **louvain_distirbution,
            **leiden_distirbution
        }

def convert_to_point(data):
    if not (hasattr(data, 'latitude') and hasattr(data, 'longitude')):
        parsed_point = data.split(" ")
        return Point(parsed_point[2][:-1:], parsed_point[1][1::])
    else:
        return data
class Point:
    def __init__(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude


from neo4j import GraphDatabase
import os
from dotenv import load_dotenv

"""
    Класс содержащий логику работы с бд neo4j
"""


class Neo4jConnection:
    def __init__(self):
        load_dotenv()

        self.__uri = os.environ.get("GRAPH_DATABASE_URL")
        self.__user = os.environ.get("GRAPH_DATABASE_USER")
        self.__pwd = os.environ.get("GRAPH_DATABASE_PASSWORD")
        self.__driver = None

        try:
            self.__driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__pwd))
        except Exception as e:
            print("Failed to create the driver:", e)

    def close(self):

        if self.__driver is not None:
            self.__driver.close()

    # TODO: need to add decorator for run and execute_write
    def run(self, query, parameters=None):
        assert self.__driver is not None, "Driver not initialized!"
        session = None

        try:
            session = self.__driver.session()
            result = session.run(query, parameters)
            print(list(result))
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()

    def execute_query(self, query, needLog=True):
        assert self.__driver is not None, "Driver not initialized!"
        session = None

        try:
            result = self.__driver.execute_query(query)
            if needLog: print(list(result))
            return result
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()

    def read_all(self, query):
        assert self.__driver is not None, "Driver not initialized!"
        session = None

        try:
            session = self.__driver.session()
            result = session.execute_read(get_node, query)
            return result
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()

    def execute_write(self, transaction_function, *args):
        assert self.__driver is not None, "Driver not initialized!"
        session = None

        try:
            session = self.__driver.session()
            session.execute_write(transaction_function, *args)
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()


def get_node(tx, query, bounds=None):
    results = tx.run(query, parameters=bounds).to_df()
    return results



from io import open
from setuptools import setup

"""
:authors: DPomortsev
:copyright: (c) 2025 DPomortsev
"""

version = "1.0.5"

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='urban_transit_network_analysis',
    version=version,

    author='DPomortsev',
    author_email='danul78969@gmail.com',

    url='https://github.com/DanilPomortsev/Urban-Transit-Network-Analysis',

    description=(
        u'Python module for urban network analysis'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=[
        "urban_transit_network_analysis",
        "urban_transit_network_analysis.context",
        "urban_transit_network_analysis.data.calculator",
        "urban_transit_network_analysis.data.preparer",
        "urban_transit_network_analysis.database",
        "urban_transit_network_analysis.enums",
        "urban_transit_network_analysis.graphics",
        "urban_transit_network_analysis.parser",
        "urban_transit_network_analysis.utils"
    ],
    install_requires=['osmnx',
                      'requests',
                      'beautifulsoup4',
                      'neo4j',
                      'pandas',
                      'plotly',
                      'deep-translator',
                      'quads',
                      'numpy',
                      'folium'
                      ]
)



from urban_transit_network_analysis.context.AnalysContext import AnalysContext
from urban_transit_network_analysis.data.calculator.MetricDataCalculator import MetricDataCalculator
from urban_transit_network_analysis.data.preparer.MetricDataPreparer import MetricDataPreparer
from urban_transit_network_analysis.database.AnalysPreparer import AnalysPreparer
from urban_transit_network_analysis.graphics.Printer import Printer
"""
    Класс ответственный за анализ сетей
"""


class AnalysManager:

    def process(self, analis_context: AnalysContext):
        ru_city_name = analis_context.ru_city_name

        for graph_analis_context in analis_context.graph_analis_context:
            graph_analis_context.city_name = ru_city_name
            db_manager_constructor = graph_analis_context.graph_type.value
            db_manager = db_manager_constructor(graph_analis_context)
            analisis_preparer = AnalysPreparer(graph_analis_context)

            if graph_analis_context.need_create_graph == True:
                db_manager.update_db(ru_city_name)
            analisis_preparer.prepare()

            prepare_result = None
            if graph_analis_context.need_prepare_data:
                metric_data_preparer = MetricDataPreparer(graph_analis_context)
                prepare_result = metric_data_preparer.prepare_metrics()

            if graph_analis_context.need_calculate_and_print_data:
                metric_data_calculator = MetricDataCalculator(graph_analis_context)
                data = metric_data_calculator.calculate_data(prepare_result)

                printer = Printer(
                    data,
                    graph_analis_context.metric_calculation_context,
                    ru_city_name
                )
                printer.print_graphics(graph_analis_context.print_graph_analis_context)



class DBGraphParameters:
    def __init__(self):
        self.main_node_name = None
        self.main_rels_name = None
        self.secondary_node_name = None
        self.secondary_rels_name = None
        self.weight = None
        self.city_name = None
        self.node_geometry_identity = None



import colorsys
import random
from enum import Enum

import plotly.express as px
import pandas as pd
import numpy as np
import plotly.graph_objects as go
import folium

from urban_transit_network_analysis.context.PrintGraphAnalysContext import PrintGraphAnalysContext
from urban_transit_network_analysis.context.MetricCalculationContext import MetricCalculationContext
"""
    Класс отрисовывающий графики по вычисленным метрикам 
"""


class Printer:
    def __init__(
            self,
            data,
            metric_calculation_context: MetricCalculationContext,
            city_name
    ):
        self.data = data
        self.metric_calculation_context = metric_calculation_context
        self.city_name = city_name

    def print_graphics(self, print_graph_analis_context: PrintGraphAnalysContext):
        for hist_metric in print_graph_analis_context.histogram_map_metrics_list:
            self.plot_histogram(hist_metric)
        for heat_map_metric in print_graph_analis_context.heat_map_metrics_list:
            self.plot_heatmap_on_map(heat_map_metric, print_graph_analis_context.mesh_size)
        for map_metric in print_graph_analis_context.map_clustering_list:
            self.plot_on_map_diff(map_metric)


    def plot_histogram(
            self,
            metric: Enum,
            title="Distribution of metric_value ",
            ylabel="Frequency"
    ):
        metric_name = str(metric.value)
        title += metric_name
        metric_values = self.data[metric_name + "_value"]

        df = pd.DataFrame({"metric_value: " + metric_name: metric_values})

        fig = px.histogram(
            df,
            x="metric_value: " + metric_name,
            title=title,
            labels={
                "metric_value ": metric_name,
                "count": ylabel
            },
            marginal="rug"
        )

        fig.show()

    def plot_heatmap_on_map(
            self,
            metric: Enum,
            resolution,
            colorscale='Viridis'
    ):
        latitudes = []
        longitudes = []
        metric_values = []
        metric_name = str(metric.value)

        for index in range(len(self.data[metric_name + "_identity"])):
            try:
                parsed_point = self.data[metric_name + "_identity"][index]
                lat, lon = parsed_point.latitude, parsed_point.longitude
                latitudes.append(float(lat))
                longitudes.append(float(lon))
                metric_values.append(self.data[metric_name + "_value"][index])
            except ValueError:
                print(f"Skipping invalid identity")

        df = pd.DataFrame({
            "latitude": latitudes,
            "longitude": longitudes,
            "metric_value": metric_values
        })
        df['metric_value'] = df['metric_value'].fillna(0)

        lat_min, lat_max = df['latitude'].min(), df['latitude'].max()
        lon_min, lon_max = df['longitude'].min(), df['longitude'].max()

        lon_edges = np.linspace(lon_min, lon_max, resolution + 1)
        lat_edges = np.linspace(lat_min, lat_max, resolution + 1)

        lon_centers = (lon_edges[:-1] + lon_edges[1:]) / 2
        lat_centers = (lat_edges[:-1] + lat_edges[1:]) / 2

        avg_values = np.empty((resolution, resolution))
        avg_values.fill(0)

        for i in range(resolution):
            for j in range(resolution):
                mask = (
                        (df['longitude'] >= lon_edges[i]) &
                        (df['longitude'] < lon_edges[i + 1]) &
                        (df['latitude'] >= lat_edges[j]) &
                        (df['latitude'] < lat_edges[j + 1])
                )

                cell_points = df[mask]

                if len(cell_points) > 0:
                    mean_value = cell_points["metric_value"].mean()
                    avg_values[j, i] = 0 if np.isnan(mean_value) or np.isinf(mean_value) else mean_value

        fig = go.Figure()

        fig.add_trace(go.Heatmap(
            z=avg_values,
            x=lon_centers,
            y=lat_centers,
            colorscale=colorscale,
            colorbar=dict(title='Среднее значение метрики ' + metric_name),
            hoverinfo='x+y+z',
            name='Средние значения ' + metric_name,
        ))

        fig.show()

    def plot_on_map(
            self,
            metric: Enum
    ):

        latitudes = []
        longitudes = []
        metric_values = []
        metric_name = str(metric.value)

        for item in range(len(self.data[metric_name + "_identity"])):
            try:
                parsed_point = self.data[metric_name + "_identity"][item]
                lat, lon = parsed_point.longitude, parsed_point.latitude
                latitudes.append(float(lat))
                longitudes.append(float(lon))
                metric_values.append(self.data[metric_name + "_value"][item])
            except ValueError:
                print(f"Skipping invalid identity")

        df = pd.DataFrame({
            "latitude": latitudes,
            "longitude": longitudes,
            "metric_value": metric_values
        })

        lat_min, lat_max = df['latitude'].min(), df['latitude'].max()
        lon_min, lon_max = df['longitude'].min(), df['longitude'].max()

        # Вычисляем центральные координаты (для центрирования карты)
        latitude = (lat_min + lat_max) / 2
        longitude = (lon_min + lon_max) / 2

        # Создаем карту, центрированную на DataFrame
        m = folium.Map(location=[longitude, latitude], zoom_start=12)

        min_metric = df['metric_value'].min()
        max_metric = df['metric_value'].max()

        # Добавьте точки на карту
        for i in range(len(df['metric_value'])):
            metric_value = df['metric_value'][i]
            radius = self.get_radius(metric_value, min_metric, max_metric)
            color = self.get_color(metric_value, min_metric, max_metric)

            folium.CircleMarker(
                location=[df['longitude'][i], df['latitude'][i]],
                radius=1,
                #color=color,
                fill=True,
                fill_color=color,
                fill_opacity=0.7,
                popup=f"Metric Value: {metric_value:.2f}",
            ).add_to(m)

        m.save(self.city_name + "_map.html")

    def plot_on_map_diff(
            self,
            metric: Enum
    ):

        latitudes = []
        longitudes = []
        metric_values = []
        metric_name = str(metric.value)

        for item in range(len(self.data[metric_name + "_identity"])):
            try:
                parsed_point = self.data[metric_name + "_identity"][item]
                lat, lon = parsed_point.longitude, parsed_point.latitude
                latitudes.append(float(lat))
                longitudes.append(float(lon))
                metric_values.append(self.data[metric_name + "_value"][item])
            except ValueError:
                print(f"Skipping invalid identity")

        df = pd.DataFrame({
            "latitude": latitudes,
            "longitude": longitudes,
            "metric_value": metric_values
        })

        grouped = df.groupby("metric_value")

        lat_min, lat_max = df['latitude'].min(), df['latitude'].max()
        lon_min, lon_max = df['longitude'].min(), df['longitude'].max()

        latitude = (lat_min + lat_max) / 2
        longitude = (lon_min + lon_max) / 2

        m = folium.Map(location=[longitude, latitude], zoom_start=12)

        for metric, group in grouped:
            color = self.get_new_color()
            coordinates = list(zip(group["latitude"], group["longitude"]))
            for point in coordinates:
                folium.CircleMarker(
                    location=[point[1], point[0]],
                    radius=1,
                    color=color,
                    fill=True,
                    fill_color=color,
                    fill_opacity=0.7,
                    popup=f"Metric Value: {metric:.2f}",
                ).add_to(m)

        m.save(self.city_name + "_" + metric_name + "_map.html")

    def get_radius(self, metric_value, min_metric, max_metric):
        normalized_value = self.normalize_metric(metric_value, min_metric, max_metric)
        min_radius = 2
        max_radius = 10
        return min_radius + (max_radius - min_radius) * normalized_value

    def get_new_color(self):
        hue = random.random()

        saturation = 0.5
        brightness = 0.8

        red, green, blue = colorsys.hsv_to_rgb(hue, saturation, brightness)

        red = int(red * 255)
        green = int(green * 255)
        blue = int(blue * 255)

        return f'#{red:02x}{green:02x}{blue:02x}'

    def normalize_metric(self, value, min_metric, max_metric):
        normalized_value = (value - min_metric) / (max_metric - min_metric)
        return normalized_value


from urban_transit_network_analysis.database.Neo4jConnection import Neo4jConnection

"""
    Класс содержащий query для вычисления метрик сети 
"""


class MetricsCalculate:
    def __init__(self, metric_name, write_property):
        self.metric_name = metric_name
        self.write_property = write_property
        self.connection = Neo4jConnection()

    def metric_calculate(
            self,
            graph_name,
            relationship_weight_property
    ):
        self.__metric_calculate(graph_name, relationship_weight_property)

    def __metric_calculate(self, graph_name, weight_property):
        query = f'''
            CALL gds.{self.metric_name}.write(
                '{graph_name}',
                {{
                relationshipWeightProperty: '{weight_property}',
                writeProperty: '{self.write_property}'
                }}
            )
        '''
        return self.connection.run(query)


class Betweenness(MetricsCalculate):
    def __init__(self):
        super().__init__("betweenness", "betweenness")


class PageRank(MetricsCalculate):
    def __init__(self):
        super().__init__("pageRank", "pageRank")
